/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  DeadFTP
 *  Copyright (C) 2000-2001 Brandon Lees
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdarg.h>
#include <dirent.h>
#include <unistd.h>
#include <sys/stat.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include <glade/glade.h>
#include <gnome-xml/tree.h>
#include <gnome-xml/parser.h>
#include <gdk-pixbuf/gdk-pixbuf.h>

#include "globals.h"
#include "main.h"
#include "ftp.h"
#include "layout.h"
#include "remote.h"
#include "queueui.h"
#include "clipboard.h"
#include "fxp.h"

/*
 * Callbacks in this file 
 */
void on_go_clicked(GtkButton * button, gpointer user_data);
void on_overwrite_clicked(GtkButton * button, gpointer user_data);
void on_resume_clicked(GtkButton * button, gpointer user_data);
void on_cancel_clicked(GtkButton * button, gpointer user_data);
void on_queue_select_row(GtkCList * clist, gint row, gint column, GdkEvent * event,
			 gpointer user_data);
void on_queue_row_move(GtkCList * clist, gint arg1, gint arg2, gpointer user_data);
void on_queue_drag_data_received(GtkWidget * widget, GdkDragContext * drag_context,
				 gint x, gint y, GtkSelectionData * data, guint info,
				 guint time, gpointer user_data);
void on_queue_selection_received(GtkWidget * widget, GtkSelectionData * data,
				 guint time, gpointer user_data);
void on_queue_unselect_row(GtkCList * clist, gint row, gint column, GdkEvent * event,
			   gpointer user_data);
void on_queue_unselect_all(GtkCList * clist, gpointer user_data);
void on_queue_file_ok_clicked(GtkButton * button, gpointer user_data);
void on_queue_ok_clicked(GtkButton * button, gpointer user_data);
void on_queue_cancel_clicked(GtkButton * button, gpointer user_data);
void on_queue_apply_clicked(GtkButton * button, gpointer user_data);
void on_editqueue_destroy(GtkObject * object, gpointer user_data);
void on_stop_clicked(GtkButton * button, gpointer user_data);
void editqueue_changed(void);
void on_editqueue_resume_toggled(GtkToggleButton * button, gpointer user_data);
gboolean on_queue_drag_drop(GtkWidget * widget, GdkDragContext * drag_context,
			    gint x, gint y, guint time);
void on_queue_clear(void);
void on_queue_move_top(void);
void on_queue_move_bottom(void);
void on_queue_move_up(void);
void on_queue_move_down(void);
void on_queue_delay(void);
void on_queue_save(void);
void on_queue_load(void);
void on_queue_edit_item(void);
void on_queue_start_item(void);
void on_queue_remove_item(void);
void on_queue_reset_status(void);

/*
 * Local function 
 */
gint queue_row_compare(gconstpointer a, gconstpointer b);
static void remove_queue(int row);
static void queue_move_item(int start, int end);
void queue_do_save(const char *filename);
void queue_do_load(const char *filename);
static void store_queue_item(queue_item_t * q);
void queue_clist_append(queue_item_t * q);
void queue_clist_insert(queue_item_t * q, gint position);
void queue_update_item_status(queue_item_t * q);
gint queue_resume_timeout(gpointer data);

void resume_dialog_destroy(void);

extern char *get_file_pixmap(const char *filename, gboolean is_dir);

GList *queue_row = NULL;
editqueue_t editqueue;
resumeDialog_t resumeDialog = { 0 };

extern GtkWidget *mainwindow;
extern GtkTargetEntry queue_targets[];

const char *queue_layout_titles[] = {
  N_("Local File"),
  N_("Local Dir"),
  N_("Remote File"),
  N_("Remote Dir"),
  N_("Hostname"),
  N_("Port"),
  N_("Username"),
  N_("Size"),
  N_("Status"),
  N_("Remote Path"),
  N_("Local Path"),
  N_("URL"),
  N_("Direction"),
};

char *queue_status_strings[] = {
  N_(""),
  N_("Paused"),
  N_("Active"),
  N_("Error"),
};

gint queue_row_compare(gconstpointer a, gconstpointer b)
{
  gint x, y;

  x = GPOINTER_TO_INT(a);
  y = GPOINTER_TO_INT(b);
  return (x - y);
}

void on_queue_move_top()
{
  GList *r, *copy;
  gint src, dst;

  queue_row = g_list_sort(queue_row, queue_row_compare);
  copy = g_list_copy(queue_row);

  /* If the first item in the queue is active, move these items to the position below so they will get
   * transferred next, otherwise move them all the way to the top */
  if (ftp_queue && QUEUE_GET_STATUS((queue_item_t *) ftp_queue->data) == QUEUE_STATUS_ACTIVE) {
    dst = 1;
  } else {
    dst = 0;
  }
  for (r = copy; r; r = g_list_next(r)) {
    src = GPOINTER_TO_INT(r->data);
    gtk_clist_row_move(GTK_CLIST(deadftp.queue), src, dst);
    dst++;
  }
  gtk_clist_unselect_all(GTK_CLIST(deadftp.queue));
  g_list_free(queue_row);
  queue_row = NULL;
  g_list_free(copy);
}

void on_queue_move_bottom()
{
  GList *r, *copy;
  gint src, dst;
  queue_item_t *q;

  queue_row = g_list_sort(queue_row, queue_row_compare);
  copy = g_list_copy(queue_row);
  for (r = copy, dst = 0; r; r = g_list_next(r)) {
    src = GPOINTER_TO_INT(r->data);
    src -= dst;
    /*
     * Need to strdup because clist_remove will free everything 
     */
    q = (queue_item_t *) g_list_nth_data(ftp_queue, src);
    gtk_clist_remove(GTK_CLIST(deadftp.queue), src);
    queue_clist_append(q);
    queue_move_item(src, g_list_length(ftp_queue));
    dst++;
  }
  gtk_clist_unselect_all(GTK_CLIST(deadftp.queue));
  g_list_free(queue_row);
  queue_row = NULL;
  g_list_free(copy);
}

void on_queue_move_up()
{
  GList *r, *copy;
  gint src;

  queue_row = g_list_sort(queue_row, queue_row_compare);
  copy = g_list_copy(queue_row);
  for (r = copy; r; r = g_list_next(r)) {
    src = GPOINTER_TO_INT(r->data);
    if (src == 0) {
      r = g_list_next(r);
      continue;
    }
    gtk_clist_row_move(GTK_CLIST(deadftp.queue), src, src - 1);
  }
  gtk_clist_unselect_all(GTK_CLIST(deadftp.queue));
  g_list_free(queue_row);
  queue_row = NULL;
  g_list_free(copy);
}

void on_queue_move_down()
{
  GList *r, *copy;
  gint src, dst;

  queue_row = g_list_sort(queue_row, queue_row_compare);
  copy = g_list_copy(queue_row);
  for (r = g_list_last(copy), dst = 1; r; r = g_list_previous(r)) {
    src = GPOINTER_TO_INT(r->data);
    gtk_clist_row_move(GTK_CLIST(deadftp.queue), src, src + 1);
  }
  gtk_clist_unselect_all(GTK_CLIST(deadftp.queue));
  g_list_free(queue_row);
  queue_row = NULL;
  g_list_free(copy);
}

void on_queue_remove_item()
{
  GList *r, *copy;
  gint src;

  queue_row = g_list_sort(queue_row, queue_row_compare);
  queue_row = g_list_reverse(queue_row);
  copy = g_list_copy(queue_row);
  for (r = copy; r; r = g_list_next(r)) {
    src = GPOINTER_TO_INT(r->data);
    remove_queue(src);
  }
  gtk_clist_unselect_all(GTK_CLIST(deadftp.queue));
  g_list_free(queue_row);
  queue_row = NULL;
  g_list_free(copy);
}

static void remove_queue(int row)
{
  queue_item_t *q;

  gtk_clist_remove(GTK_CLIST(deadftp.queue), row);
  q = g_list_nth_data(ftp_queue, row);
  ftp_queue = g_list_remove(ftp_queue, q);
  queue_item_delete(q);
}

static void queue_move_item(int start, int end)
{
  queue_item_t *q;

  q = g_list_nth_data(ftp_queue, start);
  ftp_queue = g_list_remove(ftp_queue, q);
  ftp_queue = g_list_insert(ftp_queue, q, end);
}

void queue_update_status()
{
  char buffer[REGULAR_BUFFER], c;
  unsigned long size, size_trans, file_size, t, est = 0;
  double speed;
  int id;

  g_return_if_fail(QUEUE_CURRENT(FTP.queue) != NULL);

  t = time(0);

  QUEUE_ITEM_LOCK(QUEUE_CURRENT(FTP.queue));
  /* Critical Section */
  size = QUEUE_CURRENT(FTP.queue)->size;
  size_trans = QUEUE_CURRENT(FTP.queue)->total_trans;
  file_size = QUEUE_CURRENT(FTP.queue)->total_trans;
  if (QUEUE_GET_FLAG(QUEUE_CURRENT(FTP.queue), QUEUE_FLAG_RESUME)) {
    file_size += QUEUE_CURRENT(FTP.queue)->resume;
  }

  t -= QUEUE_CURRENT(FTP.queue)->trans_start;
  /* End Critical Section */
  QUEUE_ITEM_UNLOCK(QUEUE_CURRENT(FTP.queue));

  gtk_widget_show(GTK_WIDGET(deadftp.status_progress));
  gtk_widget_show(GTK_WIDGET(deadftp.status_speed));
  gtk_widget_show(GTK_WIDGET(deadftp.status_elapsed));
  gtk_widget_show(GTK_WIDGET(deadftp.status_remaining));
  if (ftp_queue && ftp_queue->next)
    gtk_widget_show(GTK_WIDGET(deadftp.status_queue));

  gtk_progress_set_value(GTK_PROGRESS(deadftp.status_progress), file_size);

  /* status_speed */
  id =
    gtk_statusbar_get_context_id(GTK_STATUSBAR(deadftp.status_speed),
				 STATUS_CTX_SPEED);
  if (t > 3) {			/*
				 * Wait for transfer to start for more
				 * accurate speed reading 
				 */
    speed = (double) size_trans / (double) t;
    est = size - file_size;
    est /= speed;
    if (QUEUE_GET_FLAG(QUEUE_CURRENT(FTP.queue), QUEUE_FLAG_SEND))
      FTP.queue.upspeed = speed;
    else
      FTP.queue.downspeed = speed;

    c = 'B';
    if (speed > ONE_KILOBYTE) {
      speed = speed / ONE_KILOBYTE;
      c = 'K';
    }
    if (speed > ONE_KILOBYTE) {
      speed = speed / ONE_KILOBYTE;
      c = 'M';
    }
    g_snprintf(buffer, REGULAR_BUFFER, "%.2f %c/sec", speed, c);
    gtk_statusbar_pop(GTK_STATUSBAR(deadftp.status_speed), id);
    gtk_statusbar_push(GTK_STATUSBAR(deadftp.status_speed), id, buffer);
  } else {
    g_snprintf(buffer, REGULAR_BUFFER, "?? B/sec");
    gtk_statusbar_push(GTK_STATUSBAR(deadftp.status_speed), id, buffer);
  }
  /* status_elapsed */
  g_snprintf(buffer, REGULAR_BUFFER, _("Elapsed: %.2ld:%.2ld:%.2ld"), t / 3600,
	     (t % 3600) / 60, (t % 3600) % 60);
  id =
    gtk_statusbar_get_context_id(GTK_STATUSBAR(deadftp.status_elapsed),
				 STATUS_CTX_ELAPSED);
  gtk_statusbar_pop(GTK_STATUSBAR(deadftp.status_elapsed), id);
  gtk_statusbar_push(GTK_STATUSBAR(deadftp.status_elapsed), id, buffer);

  /* status_remaining */
  g_snprintf(buffer, REGULAR_BUFFER, _("Remaining: %.2ld:%.2ld:%.2ld"), est / 3600,
	     (est % 3600) / 60, (est % 3600) % 60);
  id =
    gtk_statusbar_get_context_id(GTK_STATUSBAR(deadftp.status_remaining),
				 STATUS_CTX_REMAINING);
  gtk_statusbar_pop(GTK_STATUSBAR(deadftp.status_remaining), id);
  gtk_statusbar_push(GTK_STATUSBAR(deadftp.status_remaining), id, buffer);

  /* status_queue */
  est =
    QUEUE_GET_FLAG(QUEUE_CURRENT(FTP.queue),
		   QUEUE_FLAG_SEND) ? FTP.queue.upsize -
    file_size : FTP.queue.upsize;
  t =
    QUEUE_GET_FLAG(QUEUE_CURRENT(FTP.queue),
		   QUEUE_FLAG_SEND) ? FTP.queue.downsize : FTP.queue.downsize -
    file_size;
  if (!FTP.queue.upspeed && FTP.queue.downspeed)
    FTP.queue.upspeed = FTP.queue.downspeed;
  else if (!FTP.queue.downspeed && FTP.queue.upspeed)
    FTP.queue.downspeed = FTP.queue.upspeed;

  if (est > 0)
    est /= FTP.queue.upspeed;
  if (t > 0)
    t /= FTP.queue.downspeed;

  est += t;
  if (est <= 0) {
    g_snprintf(buffer, REGULAR_BUFFER, _("Queue: ??:??:??"));
  } else {
    g_snprintf(buffer, REGULAR_BUFFER, _("Queue: %.2ld:%.2ld:%.2ld"), est / 3600,
	       (est % 3600) / 60, (est % 3600) % 60);
  }

  id =
    gtk_statusbar_get_context_id(GTK_STATUSBAR(deadftp.status_queue),
				 STATUS_CTX_QUEUE);
  gtk_statusbar_pop(GTK_STATUSBAR(deadftp.status_queue), id);
  gtk_statusbar_push(GTK_STATUSBAR(deadftp.status_queue), id, buffer);
}

void queue_pop_status()
{
  int id;

  id =
    gtk_statusbar_get_context_id(GTK_STATUSBAR(deadftp.status_speed),
				 STATUS_CTX_SPEED);
  gtk_statusbar_pop(GTK_STATUSBAR(deadftp.status_speed), id);
  gtk_statusbar_push(GTK_STATUSBAR(deadftp.status_speed), id, "");

  id =
    gtk_statusbar_get_context_id(GTK_STATUSBAR(deadftp.status_elapsed),
				 STATUS_CTX_ELAPSED);
  gtk_statusbar_pop(GTK_STATUSBAR(deadftp.status_elapsed), id);
  gtk_statusbar_push(GTK_STATUSBAR(deadftp.status_elapsed), id, "");

  id =
    gtk_statusbar_get_context_id(GTK_STATUSBAR(deadftp.status_remaining),
				 STATUS_CTX_REMAINING);
  gtk_statusbar_pop(GTK_STATUSBAR(deadftp.status_remaining), id);
  gtk_statusbar_push(GTK_STATUSBAR(deadftp.status_remaining), id, "");

  id =
    gtk_statusbar_get_context_id(GTK_STATUSBAR(deadftp.status_queue),
				 STATUS_CTX_QUEUE);
  gtk_statusbar_pop(GTK_STATUSBAR(deadftp.status_queue), id);
  gtk_statusbar_push(GTK_STATUSBAR(deadftp.status_queue), id, "");
}

void queue_hide_status()
{
  gtk_widget_hide(GTK_WIDGET(deadftp.status_progress));
  gtk_widget_hide(GTK_WIDGET(deadftp.status_speed));
  gtk_widget_hide(GTK_WIDGET(deadftp.status_elapsed));
  gtk_widget_hide(GTK_WIDGET(deadftp.status_remaining));
  gtk_widget_hide(GTK_WIDGET(deadftp.status_queue));
}

void on_queue_delay()
{
  GList *r;
  gint src;
  queue_item_t *item;

  queue_row = g_list_sort(queue_row, queue_row_compare);
  for (r = queue_row; r; r = g_list_next(r)) {
    src = GPOINTER_TO_INT(r->data);
    item = g_list_nth_data(ftp_queue, src);
    QUEUE_SET_STATUS(item, QUEUE_STATUS_PAUSE);
    queue_update_item_status(item);
  }
  gtk_clist_unselect_all(GTK_CLIST(deadftp.queue));
  g_list_free(queue_row);
}

void on_queue_save()
{
  GladeXML *xml;

  xml = glade_xml_new(deadftp.gladefile, "queue_file");
  if (!xml) {
    g_error(_("Couldn't load the interface!"));
    return;
  }

  deadftp.queueSave.dialog = glade_xml_get_widget(xml, "queue_file");
  STORE_WIDGET(deadftp.queueSave, queue_file_entry);

  /*
   * we set the data with the name "Save" to indicate that this is a save
   * dialog.  The actualy value * of the pointer is irrelevant and never
   * used 
   */
  gtk_object_set_data(GTK_OBJECT(deadftp.queueSave.dialog), "Save",
		      deadftp.queueSave.dialog);
  glade_xml_signal_autoconnect(xml);

  gtk_object_unref(GTK_OBJECT(xml));
}

void on_queue_load()
{
  GladeXML *xml;

  xml = glade_xml_new(deadftp.gladefile, "queue_file");
  if (!xml) {
    g_error(_("Couldn't load the interface!"));
    return;
  }

  deadftp.queueSave.dialog = glade_xml_get_widget(xml, "queue_file");
  STORE_WIDGET(deadftp.queueSave, queue_file_entry);

  glade_xml_signal_autoconnect(xml);

  gtk_object_unref(GTK_OBJECT(xml));
}

void queue_do_save(const char *filename)
{
  xmlDocPtr doc;
  xmlNsPtr ns;
  xmlNodePtr node;
  queue_item_t *q;
  GList *list;
  char port[20];

  g_return_if_fail(filename != NULL);

  if (!ftp_queue)
    return;
  doc = xmlNewDoc("1.0");
  doc->root = xmlNewDocNode(doc, NULL, "Queue", NULL);
  ns = xmlNewNs(doc->root, "http://deadftp.sourceforge.net/queue", "queue");

  for (list = ftp_queue; list; list = g_list_next(list)) {
    q = (queue_item_t *) list->data;

    node = xmlNewChild(doc->root, ns, "Item", NULL);
    xmlNewTextChild(node, ns, "hostname", q->hostname);

    g_snprintf(port, 20, "%d", q->port);
    xmlNewTextChild(node, ns, "port", port);

    xmlNewTextChild(node, ns, "username", q->username);
    xmlNewTextChild(node, ns, "password", q->password);
    xmlNewTextChild(node, ns, "remote_file", q->remote_file);
    xmlNewTextChild(node, ns, "remote_dir", q->remote_dir);
    xmlNewTextChild(node, ns, "local_file", q->local_file);
    xmlNewTextChild(node, ns, "local_dir", q->local_dir);

    g_snprintf(port, 20, "%d", QUEUE_GET_STATUS(q));
    xmlNewTextChild(node, ns, "status", port);
    g_snprintf(port, 20, "%ld", q->size);
    xmlNewTextChild(node, ns, "size", port);
    if (QUEUE_GET_FLAG(q, QUEUE_FLAG_SEND))
      xmlNewChild(node, ns, "send", NULL);
    if (QUEUE_GET_FLAG(q, QUEUE_FLAG_RESUME))
      xmlNewChild(node, ns, "resume", NULL);
  }
  xmlSaveFile(filename, doc);

  xmlFreeDoc(doc);
}

static void queue_add_saved_item(xmlDocPtr doc, xmlNodePtr node)
{
  xmlNodePtr sub;
  char *attr;
  queue_item_t *q;

  g_return_if_fail(doc != NULL);
  g_return_if_fail(node != NULL);

  q = g_new0(queue_item_t, 1);
  pthread_mutex_init(&q->lock, NULL);
  sub = node->childs;
  while (sub) {
    if (g_strcasecmp(sub->name, "hostname") == 0) {
      attr = xmlNodeListGetString(doc, sub->childs, 1);
      strncpy(q->hostname, attr, SMALL_BUFFER);
      g_free(attr);
    } else if (g_strcasecmp(sub->name, "port") == 0) {
      attr = xmlNodeListGetString(doc, sub->childs, 1);
      q->port = atoi(attr);
      g_free(attr);
    } else if (g_strcasecmp(sub->name, "username") == 0) {
      attr = xmlNodeListGetString(doc, sub->childs, 1);
      strncpy(q->username, attr, SMALL_BUFFER);
      g_free(attr);
    } else if (g_strcasecmp(sub->name, "password") == 0) {
      attr = xmlNodeListGetString(doc, sub->childs, 1);
      strncpy(q->password, attr, SMALL_BUFFER);
      g_free(attr);
    } else if (g_strcasecmp(sub->name, "remote_file") == 0) {
      attr = xmlNodeListGetString(doc, sub->childs, 1);
      strncpy(q->remote_file, attr, SMALL_BUFFER);
      g_free(attr);
    } else if (g_strcasecmp(sub->name, "remote_dir") == 0) {
      attr = xmlNodeListGetString(doc, sub->childs, 1);
      strncpy(q->remote_dir, attr, SMALL_BUFFER);
      g_free(attr);
    } else if (g_strcasecmp(sub->name, "local_file") == 0) {
      attr = xmlNodeListGetString(doc, sub->childs, 1);
      strncpy(q->local_file, attr, SMALL_BUFFER);
      g_free(attr);
    } else if (g_strcasecmp(sub->name, "local_dir") == 0) {
      attr = xmlNodeListGetString(doc, sub->childs, 1);
      strncpy(q->local_dir, attr, SMALL_BUFFER);
      g_free(attr);
    } else if (g_strcasecmp(sub->name, "status") == 0) {
      attr = xmlNodeListGetString(doc, sub->childs, 1);
      QUEUE_SET_STATUS(q, atoi(attr));
      if (QUEUE_GET_STATUS(q) == QUEUE_STATUS_ACTIVE)
	QUEUE_SET_STATUS(q, QUEUE_STATUS_NORMAL);
      g_free(attr);
    } else if (g_strcasecmp(sub->name, "size") == 0) {
      attr = xmlNodeListGetString(doc, sub->childs, 1);
      q->size = atoi(attr);
      g_free(attr);
    } else if (g_strcasecmp(sub->name, "send") == 0) {
      QUEUE_SET_FLAG(q, QUEUE_FLAG_SEND);
    } else if (g_strcasecmp(sub->name, "resume") == 0) {
      QUEUE_SET_FLAG(q, QUEUE_FLAG_RESUME);
    }
    sub = sub->next;
  }
  add_queue_item(q);
}

void queue_do_load(const char *filename)
{
  FILE *f;
  xmlDocPtr doc;
  xmlNsPtr ns;
  xmlNodePtr node;

  g_return_if_fail(filename != NULL);

  f = fopen(filename, "r");
  if (f != NULL) {
    int res, size = REGULAR_BUFFER;
    char chars[REGULAR_BUFFER];
    xmlParserCtxtPtr ctxt = NULL;

    res = fread(chars, 1, 4, f);
    if (res > 0) {
      ctxt = xmlCreatePushParserCtxt(NULL, NULL, chars, res, filename);
      while ((res = fread(chars, 1, size, f)) > 0) {
	xmlParseChunk(ctxt, chars, res, 0);
      }
    }
    xmlParseChunk(ctxt, chars, 0, 1);
    doc = ctxt->myDoc;
    xmlFreeParserCtxt(ctxt);
    fclose(f);
  } else {
    return;
  }

  if (!doc || !doc->root || (g_strcasecmp(doc->root->name, "Queue") != 0)) {
    /*
     * Invalid File.  (Display dialog) 
     */
    return;
  }
  ns = xmlSearchNsByHref(doc, doc->root, "http://deadftp.sourceforge.net/queue");
  if (!ns) {
    /*
     * Inavlid File. (Display dialog) 
     */
    return;
  }

  node = doc->root->childs;
  while (node) {
    if (g_strcasecmp(node->name, "Item") == 0) {
      queue_add_saved_item(doc, node);
    }
    node = node->next;
  }
  xmlFreeDoc(doc);
}

void on_go_clicked(GtkButton * button, gpointer user_data)
{
  if (FTP.state == FTP_STATE_IDLE || FTP.state == FTP_STATE_DISCONNECTED) {
    /* Only go if we aren't doing anything */
    queue_start();
  }
}

void on_queue_select_row(GtkCList * clist, gint row, gint column, GdkEvent * event,
			 gpointer user_data)
{
  if (!row_select_find(queue_row, row)) {
    queue_row = g_list_append(queue_row, GINT_TO_POINTER(row));
  }
}

void on_queue_row_move(GtkCList * clist, gint arg1, gint arg2, gpointer user_data)
{
  queue_move_item(arg1, arg2);
  gtk_clist_unselect_all(GTK_CLIST(deadftp.queue));
  g_list_free(queue_row);
  queue_row = NULL;
}

void on_queue_drag_data_received(GtkWidget * widget, GdkDragContext * drag_context,
				 gint x, gint y, GtkSelectionData * data, guint info,
				 guint time, gpointer user_data)
{
  if (data->target == gdk_atom_intern("xml/fxp-info", FALSE)) {
    xmlDocPtr doc = xmlParseMemory(data->data, data->length - 1);
    xmlNodePtr node, sub;
    fxp_desc_t *fxp = NULL;

    for (node = doc->root->childs; node; node = node->next) {
      if (g_strcasecmp(node->name, "fxp") == 0) {
	for (sub = node->childs; sub; sub = sub->next) {
	  if (g_strcasecmp(sub->name, "shmid") == 0) {
	    xmlChar *shmid = xmlNodeListGetString(doc, sub->childs, TRUE);
	    fxp = fxp_attach(atoi(shmid));
	    g_free(shmid);
	  }
	}
      } else if (g_strcasecmp(node->name, "item") == 0) {
	xmlChar *hostname = NULL, *filename = NULL, *length = NULL, *qid = NULL;
	for (sub = node->childs; sub; sub = sub->next) {
	  if (g_strcasecmp(sub->name, "hostname") == 0)
	    hostname = xmlNodeListGetString(doc, sub->childs, TRUE);
	  else if (g_strcasecmp(sub->name, "filename") == 0)
	    filename = xmlNodeListGetString(doc, sub->childs, TRUE);
	  else if (g_strcasecmp(sub->name, "length") == 0)
	    length = xmlNodeListGetString(doc, sub->childs, TRUE);
	  else if (g_strcasecmp(sub->name, "id") == 0)
	    qid = xmlNodeListGetString(doc, sub->childs, TRUE);
	}
	queue_add_fxp(fxp, FTP.server.hostname, FTP.server.port, FTP.server.username,
		      FTP.server.password, FTP.server.current_directory, filename,
		      length ? atoi(length) : 0, qid ? atoi(qid) : 0);
	if (hostname)
	  g_free(hostname);
	if (filename)
	  g_free(filename);
	if (length)
	  g_free(length);
      }
    }
    xmlFreeDoc(doc);
  } else if (data->target == gdk_atom_intern("xml/file-info", FALSE)) {
    xmlDocPtr doc = xmlParseMemory(data->data, data->length - 1);
    xmlNodePtr node, sub;

    for (node = doc->root->childs; node; node = node->next) {
      xmlChar *local_dir = NULL, *local_file = NULL, *cp;
      gboolean send = FALSE;
      ftp_server_t server;
      struct afile file;

      bzero((char *) &server, sizeof(server));
      bzero((char *) &file, sizeof(file));

      if (g_strcasecmp(node->name, "item") != 0) {
	g_warning("Got invalid tag in dropped file-info: %s\n", node->name);
	node = node->next;
      }

      for (sub = node->childs; sub; sub = sub->next) {
	if (g_strcasecmp(sub->name, "hostname") == 0) {
	  cp = xmlNodeListGetString(doc, sub->childs, TRUE);
	  strncpy(server.hostname, cp, SMALL_BUFFER);
	  g_free(cp);
	} else if (g_strcasecmp(sub->name, "port") == 0) {
	  cp = xmlNodeListGetString(doc, sub->childs, TRUE);
	  server.port = atoi(cp);
	  g_free(cp);
	} else if (g_strcasecmp(sub->name, "username") == 0) {
	  cp = xmlNodeListGetString(doc, sub->childs, TRUE);
	  strncpy(server.username, cp, SMALL_BUFFER);
	  g_free(cp);
	} else if (g_strcasecmp(sub->name, "password") == 0) {
	  cp = xmlNodeListGetString(doc, sub->childs, TRUE);
	  strncpy(server.password, cp, SMALL_BUFFER);
	  g_free(cp);
	} else if (g_strcasecmp(sub->name, "remote_file") == 0) {
	  file.file = xmlNodeListGetString(doc, sub->childs, TRUE);
	} else if (g_strcasecmp(sub->name, "remote_dir") == 0) {
	  cp = xmlNodeListGetString(doc, sub->childs, TRUE);
	  strncpy(server.current_directory, cp, REGULAR_BUFFER);
	  g_free(cp);
	} else if (g_strcasecmp(sub->name, "local_file") == 0) {
	  local_file = xmlNodeListGetString(doc, sub->childs, TRUE);
	} else if (g_strcasecmp(sub->name, "local_dir") == 0) {
	  local_dir = xmlNodeListGetString(doc, sub->childs, TRUE);
	} else if (g_strcasecmp(sub->name, "length") == 0) {
	  file.length = xmlNodeListGetString(doc, sub->childs, TRUE);
	} else if (g_strcasecmp(sub->name, "send") == 0) {
	  send = TRUE;
	} else if (g_strcasecmp(sub->name, "directory") == 0) {
	  file.is_dir = TRUE;
	} else if (g_strcasecmp(sub->name, "date") == 0) {
	  file.date = xmlNodeListGetString(doc, sub->childs, TRUE);
	}
      }
      if (!server.port)
	server.port = 21;
      /* If this is a send we have to move local_file into the file struct
       * so it's copied to the queue_item in the correct position to be the
       * source file and remote_file into dest_file */
      if (send) {
	cp = g_strdup(file.file);
	strncpy(file.file, local_file, SMALL_BUFFER);
	g_free(local_file);
	local_file = cp;
      }
      /* make a queue item out of the information */
      add_queue(&server, &file, local_dir, local_file, send);

      /* free memory */
      if (local_file)
	g_free(local_file);
      if (local_dir)
	g_free(local_dir);
      if (file.file)
	g_free(file.file);
      if (file.length)
	g_free(file.length);
    }
    /* Free the xml document */
    xmlFreeDoc(doc);
  } else {
    /* make sure the data is null terminated */
    data->data[data->length] = '\0';
    queue_add_url(data->data);
  }
}

void on_queue_selection_received(GtkWidget * widget, GtkSelectionData * data,
				 guint time, gpointer user_data)
{
  char **lists, buffer[REGULAR_BUFFER];
  gint count, i;
  enum { INVALID, STRING, CTEXT } type;

  if (data->type == GDK_TARGET_STRING)
    type = STRING;
  else if ((data->type == gdk_atom_intern("COMPOUND_TEXT", FALSE))
	   || (data->type == gdk_atom_intern("TEXT", FALSE)))
    type = CTEXT;
  else
    type = INVALID;

  if (type == INVALID || data->length < 0) {
    if (data->target != GDK_TARGET_STRING)
      gtk_selection_convert(widget, data->selection, GDK_TARGET_STRING, time);
  }
  switch (type) {
  case STRING:
    data->data[data->length] = '\0';
    strncpy(buffer, data->data, REGULAR_BUFFER);
    break;
  case CTEXT:
    count =
      gdk_text_property_to_text_list(data->type, data->format, data->data,
				     data->length, &lists);
    *buffer = '\0';
    for (i = 0; i < count; i++)
      strncat(buffer, lists[i], REGULAR_BUFFER);
    if (count)
      gdk_free_text_list(lists);
    break;
  case INVALID:
    break;
  }

  if (clipboard_next_convert && deadftp_prefs.clipboard_monitor) {
    gtk_selection_convert(widget, gdk_atom_intern("CLIPBOARD", FALSE),
			  gdk_atom_intern("COMPOUND_TEXT", FALSE), GDK_CURRENT_TIME);
  }

  if (type == INVALID || data->length < 0)
    return;

  if (data->selection == GDK_SELECTION_PRIMARY) {
    if (!have_selection) {
      strncpy(selection_text, buffer, REGULAR_BUFFER);
      have_selection = 1;
      return;
    }
    if (g_strcasecmp(selection_text, buffer) == 0)
      return;
    strncpy(selection_text, buffer, REGULAR_BUFFER);
  } else {
    clipboard_next_convert = 0;
    if (!have_clipboard) {
      strncpy(clipboard_text, buffer, REGULAR_BUFFER);
      have_clipboard = 1;
      return;
    }
    if (g_strcasecmp(clipboard_text, buffer) == 0)
      return;
    strncpy(clipboard_text, buffer, REGULAR_BUFFER);
  }

  queue_add_url(buffer);
}

void on_queue_unselect_row(GtkCList * clist, gint row, gint column, GdkEvent * event,
			   gpointer user_data)
{
  GList *x = row_select_find(queue_row, row);

  if (!x)
    return;
  queue_row = g_list_remove(queue_row, x->data);

}

void on_queue_unselect_all(GtkCList * clist, gpointer user_data)
{
  /* Never really seen this called */
  g_print("unselect all for queue\n");
}

void on_queue_file_ok_clicked(GtkButton * button, gpointer user_data)
{
  char *filename;

  if (gtk_object_get_data(GTK_OBJECT(deadftp.queueSave.dialog), "Save")) {
    filename =
      gnome_file_entry_get_full_path(GNOME_FILE_ENTRY
				     (deadftp.queueSave.queue_file_entry), FALSE);
    queue_do_save(filename);
    g_free(filename);
  } else {
    filename =
      gnome_file_entry_get_full_path(GNOME_FILE_ENTRY
				     (deadftp.queueSave.queue_file_entry), TRUE);
    queue_do_load(filename);
    g_free(filename);
  }
}

static void load_edit_queue(queue_item_t * q)
{
  GladeXML *xml;
  GList *item;
  int i;

  xml = glade_xml_new(deadftp.gladefile, "editqueue");
  if (!xml) {
    g_error(_("Couldn't load the interface!"));
    return;
  }

  editqueue.window = glade_xml_get_widget(xml, "editqueue");

  STORE_WIDGET(editqueue, hostname);
  STORE_WIDGET(editqueue, port);
  STORE_WIDGET(editqueue, username);
  STORE_WIDGET(editqueue, password);

  STORE_WIDGET(editqueue, local_dir);
  STORE_WIDGET(editqueue, local_file);
  STORE_WIDGET(editqueue, remote_dir);
  STORE_WIDGET(editqueue, remote_file);

  STORE_WIDGET(editqueue, resume);
  STORE_WIDGET(editqueue, resume_place);
  STORE_WIDGET(editqueue, status);
  STORE_WIDGET(editqueue, status_entry);

  STORE_WIDGET(editqueue, queue_apply);

  editqueue.changed = FALSE;
  gtk_widget_set_sensitive(GTK_WIDGET(editqueue.queue_apply), FALSE);

  /* Queue item we're editing */
  gtk_object_set_data(GTK_OBJECT(editqueue.window), "queue_item", (gpointer) q);

  gtk_entry_set_text(GTK_ENTRY(editqueue.hostname), q->hostname);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(editqueue.port), (float) q->port);
  gtk_entry_set_text(GTK_ENTRY(editqueue.username), q->username);
  gtk_entry_set_text(GTK_ENTRY(editqueue.password), q->password);

  gtk_entry_set_text(GTK_ENTRY(editqueue.local_dir), q->local_dir);
  gtk_entry_set_text(GTK_ENTRY(editqueue.local_file), q->local_file);
  gtk_entry_set_text(GTK_ENTRY(editqueue.remote_dir), q->remote_dir);
  gtk_entry_set_text(GTK_ENTRY(editqueue.remote_file), q->remote_file);

  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(editqueue.resume),
			       QUEUE_GET_FLAG(q, QUEUE_FLAG_RESUME));
  gtk_widget_set_sensitive(GTK_WIDGET(editqueue.resume_place),
			   QUEUE_GET_FLAG(q, QUEUE_FLAG_RESUME));
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(editqueue.resume_place), q->resume);

  for (i = 0, item = NULL; i < NUM_QUEUE_STATUS; i++) {
    item = g_list_append(item, _(queue_status_strings[i]));
  }
  gtk_combo_set_popdown_strings(GTK_COMBO(editqueue.status), item);
  gtk_entry_set_text(GTK_ENTRY(editqueue.status_entry),
		     _(queue_status_strings[QUEUE_GET_STATUS(q)]));

  g_list_free(item);

  glade_xml_signal_autoconnect(xml);

  gtk_object_unref(GTK_OBJECT(xml));
}

static void store_queue_item(queue_item_t * q)
{
  char *value;
  int i;

  g_return_if_fail(q != NULL);

  value = gtk_entry_get_text(GTK_ENTRY(editqueue.hostname));
  strncpy(q->hostname, value, SMALL_BUFFER);

  q->port = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(editqueue.port));

  value = gtk_entry_get_text(GTK_ENTRY(editqueue.username));
  strncpy(q->username, value, SMALL_BUFFER);

  value = gtk_entry_get_text(GTK_ENTRY(editqueue.password));
  strncpy(q->password, value, SMALL_BUFFER);

  value = gtk_entry_get_text(GTK_ENTRY(editqueue.local_dir));
  strncpy(q->local_dir, value, SMALL_BUFFER);

  value = gtk_entry_get_text(GTK_ENTRY(editqueue.local_file));
  strncpy(q->local_file, value, SMALL_BUFFER);

  value = gtk_entry_get_text(GTK_ENTRY(editqueue.remote_dir));
  strncpy(q->remote_dir, value, SMALL_BUFFER);

  value = gtk_entry_get_text(GTK_ENTRY(editqueue.remote_file));
  strncpy(q->remote_file, value, SMALL_BUFFER);

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(editqueue.resume))) {
    QUEUE_SET_FLAG(q, QUEUE_FLAG_RESUME);
  } else {
    QUEUE_CLEAR_FLAG(q, QUEUE_FLAG_RESUME);
  }
  q->resume =
    gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(editqueue.resume_place));

  value = gtk_entry_get_text(GTK_ENTRY(editqueue.status_entry));
  QUEUE_SET_STATUS(q, QUEUE_STATUS_NORMAL);
  i = 0;
  while ((i < NUM_QUEUE_STATUS)
	 && (g_strcasecmp(value, _(queue_status_strings[i])) != 0)) {
    i++;
  }
  if (i < NUM_QUEUE_STATUS) {
    QUEUE_SET_STATUS(q, i);
  }
  queue_update_item_status(q);
}

void on_queue_ok_clicked(GtkButton * button, gpointer user_data)
{
  queue_item_t *q;

  q =
    (queue_item_t *) gtk_object_get_data(GTK_OBJECT(editqueue.window), "queue_item");
  g_return_if_fail(q != NULL);

  if (editqueue.changed) {
    store_queue_item(q);
  }

  gtk_widget_destroy(editqueue.window);
}

void on_queue_cancel_clicked(GtkButton * button, gpointer user_data)
{
  gtk_widget_destroy(editqueue.window);
}

void on_queue_apply_clicked(GtkButton * button, gpointer user_data)
{
  queue_item_t *q;

  q =
    (queue_item_t *) gtk_object_get_data(GTK_OBJECT(editqueue.window), "queue_item");
  g_return_if_fail(q != NULL);

  store_queue_item(q);

  editqueue.changed = FALSE;
  gtk_widget_set_sensitive(GTK_WIDGET(editqueue.queue_apply), FALSE);
}

void on_editqueue_destroy(GtkObject * object, gpointer user_data)
{
  gtk_widget_destroy(editqueue.window);
}

void on_queue_edit_item()
{
  queue_item_t *q;

  if (!queue_row)
    return;

  q = (queue_item_t *) g_list_nth_data(ftp_queue, GPOINTER_TO_INT(queue_row->data));
  load_edit_queue(q);
}

void init_queue_layout()
{
  int i;

  gtk_clist_freeze(GTK_CLIST(deadftp.queue));

  for (i = 1; i < NUM_QUEUE_LAYOUTS; i++) {
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.queue), i, FALSE);
  }

  if (deadftp_prefs.queue_layout[DIR_LAYOUT_FILE] != -1) {
    gtk_clist_set_column_title(GTK_CLIST(deadftp.queue),
			       deadftp_prefs.queue_layout[QUEUE_LAYOUT_LOCAL_FILE],
			       _(queue_layout_titles[QUEUE_LAYOUT_LOCAL_FILE]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.queue),
				    deadftp_prefs.
				    queue_layout[QUEUE_LAYOUT_LOCAL_FILE], TRUE);
  }
  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_LOCAL_DIR] != -1) {
    gtk_clist_set_column_title(GTK_CLIST(deadftp.queue),
			       deadftp_prefs.queue_layout[QUEUE_LAYOUT_LOCAL_DIR],
			       _(queue_layout_titles[QUEUE_LAYOUT_LOCAL_DIR]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.queue),
				    deadftp_prefs.
				    queue_layout[QUEUE_LAYOUT_LOCAL_DIR], TRUE);
  }
  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_REMOTE_FILE] != -1) {
    gtk_clist_set_column_title(GTK_CLIST(deadftp.queue),
			       deadftp_prefs.queue_layout[QUEUE_LAYOUT_REMOTE_FILE],
			       _(queue_layout_titles[QUEUE_LAYOUT_REMOTE_FILE]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.queue),
				    deadftp_prefs.
				    queue_layout[QUEUE_LAYOUT_REMOTE_FILE], TRUE);
  }
  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_REMOTE_DIR] != -1) {
    gtk_clist_set_column_title(GTK_CLIST(deadftp.queue),
			       deadftp_prefs.queue_layout[QUEUE_LAYOUT_REMOTE_DIR],
			       _(queue_layout_titles[QUEUE_LAYOUT_REMOTE_DIR]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.queue),
				    deadftp_prefs.
				    queue_layout[QUEUE_LAYOUT_REMOTE_DIR], TRUE);
  }
  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_HOSTNAME] != -1) {
    gtk_clist_set_column_title(GTK_CLIST(deadftp.queue),
			       deadftp_prefs.queue_layout[QUEUE_LAYOUT_HOSTNAME],
			       _(queue_layout_titles[QUEUE_LAYOUT_HOSTNAME]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.queue),
				    deadftp_prefs.
				    queue_layout[QUEUE_LAYOUT_HOSTNAME], TRUE);
  }
  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_PORT] != -1) {
    gtk_clist_set_column_title(GTK_CLIST(deadftp.queue),
			       deadftp_prefs.queue_layout[QUEUE_LAYOUT_PORT],
			       _(queue_layout_titles[QUEUE_LAYOUT_PORT]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.queue),
				    deadftp_prefs.queue_layout[QUEUE_LAYOUT_PORT],
				    TRUE);
  }
  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_USERNAME] != -1) {
    gtk_clist_set_column_title(GTK_CLIST(deadftp.queue),
			       deadftp_prefs.queue_layout[QUEUE_LAYOUT_USERNAME],
			       _(queue_layout_titles[QUEUE_LAYOUT_USERNAME]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.queue),
				    deadftp_prefs.
				    queue_layout[QUEUE_LAYOUT_USERNAME], TRUE);
  }
  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_SIZE] != -1) {
    gtk_clist_set_column_title(GTK_CLIST(deadftp.queue),
			       deadftp_prefs.queue_layout[QUEUE_LAYOUT_SIZE],
			       _(queue_layout_titles[QUEUE_LAYOUT_SIZE]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.queue),
				    deadftp_prefs.queue_layout[QUEUE_LAYOUT_SIZE],
				    TRUE);
  }
  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_STATUS] != -1) {
    gtk_clist_set_column_title(GTK_CLIST(deadftp.queue),
			       deadftp_prefs.queue_layout[QUEUE_LAYOUT_STATUS],
			       _(queue_layout_titles[QUEUE_LAYOUT_STATUS]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.queue),
				    deadftp_prefs.queue_layout[QUEUE_LAYOUT_STATUS],
				    TRUE);
  }
  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_REMOTE_PATH] != -1) {
    gtk_clist_set_column_title(GTK_CLIST(deadftp.queue),
			       deadftp_prefs.queue_layout[QUEUE_LAYOUT_REMOTE_PATH],
			       _(queue_layout_titles[QUEUE_LAYOUT_REMOTE_PATH]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.queue),
				    deadftp_prefs.
				    queue_layout[QUEUE_LAYOUT_REMOTE_PATH], TRUE);
  }
  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_LOCAL_PATH] != -1) {
    gtk_clist_set_column_title(GTK_CLIST(deadftp.queue),
			       deadftp_prefs.queue_layout[QUEUE_LAYOUT_LOCAL_PATH],
			       _(queue_layout_titles[QUEUE_LAYOUT_LOCAL_PATH]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.queue),
				    deadftp_prefs.
				    queue_layout[QUEUE_LAYOUT_LOCAL_PATH], TRUE);
  }
  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_URL] != -1) {
    gtk_clist_set_column_title(GTK_CLIST(deadftp.queue),
			       deadftp_prefs.queue_layout[QUEUE_LAYOUT_URL],
			       _(queue_layout_titles[QUEUE_LAYOUT_URL]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.queue),
				    deadftp_prefs.queue_layout[QUEUE_LAYOUT_URL],
				    TRUE);
  }
  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_DIRECTION] != -1) {
    gtk_clist_set_column_title(GTK_CLIST(deadftp.queue),
			       deadftp_prefs.queue_layout[QUEUE_LAYOUT_DIRECTION],
			       _(queue_layout_titles[QUEUE_LAYOUT_DIRECTION]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.queue),
				    deadftp_prefs.
				    queue_layout[QUEUE_LAYOUT_DIRECTION], TRUE);
  }


  for (i = 0; i < NUM_QUEUE_LAYOUTS; i++) {
    gtk_clist_set_column_auto_resize(GTK_CLIST(deadftp.queue), i, TRUE);
  }

  gtk_clist_thaw(GTK_CLIST(deadftp.queue));
}

void queue_clist_append(queue_item_t * q)
{
  queue_clist_insert(q, g_list_length(ftp_queue) + 1);
}

void queue_clist_insert(queue_item_t * q, gint row)
{
  char *qent[NUM_QUEUE_LAYOUTS], size[SMALL_BUFFER], remote_path[REGULAR_BUFFER];
  char local_path[REGULAR_BUFFER], url[REGULAR_BUFFER], port[SMALL_BUFFER];
  int i;

  g_return_if_fail(q != NULL);

  for (i = 0; i < NUM_QUEUE_LAYOUTS; i++)
    qent[i] = NULL;

  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_LOCAL_FILE] != -1)
    qent[deadftp_prefs.queue_layout[QUEUE_LAYOUT_LOCAL_FILE]] = q->local_file;

  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_LOCAL_DIR] != -1)
    qent[deadftp_prefs.queue_layout[QUEUE_LAYOUT_LOCAL_DIR]] = q->local_dir;

  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_REMOTE_FILE] != -1)
    qent[deadftp_prefs.queue_layout[QUEUE_LAYOUT_REMOTE_FILE]] = q->remote_file;

  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_REMOTE_DIR] != -1)
    qent[deadftp_prefs.queue_layout[QUEUE_LAYOUT_REMOTE_DIR]] = q->remote_dir;

  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_HOSTNAME] != -1)
    qent[deadftp_prefs.queue_layout[QUEUE_LAYOUT_HOSTNAME]] = q->hostname;

  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_PORT] != -1) {
    g_snprintf(port, SMALL_BUFFER, "%d", q->port);
    qent[deadftp_prefs.queue_layout[QUEUE_LAYOUT_PORT]] = port;
  }

  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_USERNAME] != -1)
    qent[deadftp_prefs.queue_layout[QUEUE_LAYOUT_USERNAME]] = q->username;

  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_SIZE] != -1) {
    g_snprintf(size, SMALL_BUFFER, "%ld", q->size);
    qent[deadftp_prefs.queue_layout[QUEUE_LAYOUT_SIZE]] = size;
  }

  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_STATUS] != -1)
    qent[deadftp_prefs.queue_layout[QUEUE_LAYOUT_STATUS]] =
      _(queue_status_strings[QUEUE_GET_STATUS(q)]);

  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_REMOTE_PATH] != -1) {
    g_snprintf(remote_path, REGULAR_BUFFER, "%s/%s", q->remote_dir, q->remote_file);
    qent[deadftp_prefs.queue_layout[QUEUE_LAYOUT_REMOTE_PATH]] = remote_path;
  }

  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_LOCAL_PATH] != -1) {
    g_snprintf(local_path, REGULAR_BUFFER, "%s/%s", q->local_dir, q->local_file);
    qent[deadftp_prefs.queue_layout[QUEUE_LAYOUT_LOCAL_PATH]] = local_path;
  }

  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_URL] != -1) {
    /* TODO: write a function that creates the URL since we need to do it other places as well */
    g_snprintf(port, SMALL_BUFFER, "%d", q->port);
    if (g_strcasecmp(q->username, "anonymous") == 0) {
      g_snprintf(url, REGULAR_BUFFER, "ftp://%s%s%s%s/%s", q->hostname,
		 (q->port != 21) ? ":" : "", (q->port != 21) ? port : "",
		 q->remote_dir, q->remote_file);
    } else {
      g_snprintf(url, REGULAR_BUFFER, "ftp://%s@%s%s%s%s/%s", q->username,
		 q->hostname, (q->port != 21) ? ":" : "",
		 (q->port != 21) ? port : "", q->remote_dir, q->remote_file);
    }
    qent[deadftp_prefs.queue_layout[QUEUE_LAYOUT_URL]] = url;
  }
  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_DIRECTION] != -1) {
    if (q->fxp) {
      qent[deadftp_prefs.queue_layout[QUEUE_LAYOUT_DIRECTION]] = "FXP";
    } else if (QUEUE_GET_FLAG(q, QUEUE_FLAG_SEND)) {
      qent[deadftp_prefs.queue_layout[QUEUE_LAYOUT_DIRECTION]] = "->";
    } else {
      qent[deadftp_prefs.queue_layout[QUEUE_LAYOUT_DIRECTION]] = "<-";
    }
  }

  row = gtk_clist_insert(GTK_CLIST(deadftp.queue), row, qent);

  set_clist_icon(deadftp.queue, row, deadftp_prefs.directory_layout[DIR_LAYOUT_FILE],
		 q->local_file, QUEUE_GET_FLAG(q, QUEUE_FLAG_DIRECTORY));
}

void redisplay_queue()
{
  GList *l;
  queue_item_t *q;

  for (l = ftp_queue; l; l = g_list_next(l)) {
    q = (queue_item_t *) l->data;

    queue_clist_append(q);
  }
}

void on_stop_clicked(GtkButton * button, gpointer user_data)
{
  queue_item_t *q;

  if (FTP.state == FTP_STATE_IDLE) {
    /* If we're not doing anything just return */
    return;
  }

  if (ftp_queue_current) {
    q = (queue_item_t *) ftp_queue_current->data;
    QUEUE_SET_STATUS(q, QUEUE_STATUS_NORMAL);
    queue_update_item_status(q);
  }

  if (FTP.state == FTP_STATE_TRANS_RETR) {
    ftp_abort();
  }

  ftp_clear_commands();

  if (FTP.retry_tag != -1) {
    gtk_timeout_remove(FTP.retry_tag);
    FTP.retry_tag = -1;
    print_console(CONSOLE_INFO, _("Retry Aborted\n"));
  }
  /* Send a ABOR so that we get a reply from the server which we can use
   * to determine what replires where left over from other commands
   */
  ftp_send_command(FTP_STATE_ABORT, 0, _("Aborting"), "ABOR");
}

void queue_update_item_status(queue_item_t * q)
{
  g_return_if_fail(q != NULL);

  if (QUEUE_GET_STATUS(q) < 0 || QUEUE_GET_STATUS(q) >= NUM_QUEUE_STATUS) {
    return;
  }
  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_STATUS] != -1) {
    gtk_clist_set_text(GTK_CLIST(deadftp.queue), g_list_index(ftp_queue, q),
		       deadftp_prefs.queue_layout[QUEUE_LAYOUT_STATUS],
		       _(queue_status_strings[QUEUE_GET_STATUS(q)]));
  }
}

void on_queue_reset_status(void)
{
  GList *r;
  gint src;
  queue_item_t *q;

  queue_row = g_list_sort(queue_row, queue_row_compare);
  for (r = queue_row; r; r = g_list_next(r)) {
    src = GPOINTER_TO_INT(r->data);
    q = (queue_item_t *) g_list_nth_data(ftp_queue, src);

    QUEUE_SET_STATUS(q, QUEUE_STATUS_NORMAL);
    queue_update_item_status(q);
  }
}

void editqueue_changed(void)
{
  editqueue.changed = TRUE;
  gtk_widget_set_sensitive(GTK_WIDGET(editqueue.queue_apply), TRUE);
}

void on_editqueue_resume_toggled(GtkToggleButton * button, gpointer user_data)
{
  queue_item_t *q;

  q =
    (queue_item_t *) gtk_object_get_data(GTK_OBJECT(editqueue.window), "queue_item");
  g_return_if_fail(q != NULL);

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(editqueue.resume))) {
    QUEUE_SET_FLAG(q, QUEUE_FLAG_RESUME);
  } else {
    QUEUE_CLEAR_FLAG(q, QUEUE_FLAG_RESUME);
  }
  gtk_widget_set_sensitive(GTK_WIDGET(editqueue.resume_place),
			   QUEUE_GET_FLAG(q, QUEUE_FLAG_RESUME));

  editqueue_changed();
}

void on_queue_start_item()
{
  GList *r;
  int src;

  if (FTP.state != FTP_STATE_IDLE)
    return;

  if (!queue_row)
    return;

  queue_row = g_list_sort(queue_row, queue_row_compare);
  src = GPOINTER_TO_INT(queue_row->data);
  r = g_list_nth(ftp_queue, src);
  if (r)
    queue_start_item((queue_item_t *) r->data);
}


gboolean on_queue_drag_drop(GtkWidget * widget, GdkDragContext * drag_context,
			    gint x, gint y, guint time)
{
  GList *tmp_source = NULL;
  GtkTargetEntry *target;
  int i;
  GtkWidget *source_widget = gtk_drag_get_source_widget(drag_context);
  GdkAtom atom = gdk_atom_intern("xml/fxp-info", FALSE);

  for (tmp_source = drag_context->targets; tmp_source && !source_widget;
       tmp_source = tmp_source->next) {
    if (atom == GPOINTER_TO_INT(tmp_source->data)) {
      gtk_drag_get_data(widget, drag_context, atom, time);
      return TRUE;
    }
  }

  target = queue_targets;
  i = 0;
  while (i < 6) {
    atom = gdk_atom_intern(target->target, FALSE);
    tmp_source = drag_context->targets;
    while (tmp_source) {
      if (atom == GPOINTER_TO_INT(tmp_source->data)) {
	if ((!(target->flags & GTK_TARGET_SAME_APP) || source_widget)
	    && (!(target->flags & GTK_TARGET_SAME_WIDGET)
		|| (source_widget == widget))) {

	  gtk_drag_get_data(widget, drag_context, atom, time);
	  return TRUE;
	} else
	  break;
      }
      tmp_source = tmp_source->next;
    }
    target++;
    i++;
  }
  return FALSE;
}

void on_queue_clear()
{
  GList *l;

  gtk_clist_clear(GTK_CLIST(deadftp.queue));
  for (l = ftp_queue; l; l = g_list_next(l))
    queue_item_delete((queue_item_t *) l->data);
  g_list_free(ftp_queue);
  ftp_queue = NULL;
}

void queue_update_item_size(queue_item_t * q)
{
  g_return_if_fail(q != NULL);

  if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_STATUS] != -1) {
    char size[32];
    g_snprintf(size, 32, "%ld", q->size);
    gtk_clist_set_text(GTK_CLIST(deadftp.queue), g_list_index(ftp_queue, q),
		       deadftp_prefs.queue_layout[QUEUE_LAYOUT_SIZE], size);
  }
}

static void set_file_pixmap(GtkPixmap * pixmap, const char *filename)
{
  char *pixfile;

  pixfile = get_file_pixmap(filename, FALSE);
  if (pixfile) {
    GdkPixbuf *pb = gdk_pixbuf_new_from_file(pixfile);
    GdkPixmap *pm = NULL;
    GdkBitmap *bm = NULL;

    gdk_pixbuf_render_pixmap_and_mask(pb, &pm, &bm, 128);
    gtk_pixmap_set(GTK_PIXMAP(pixmap), pm, bm);

    gdk_pixbuf_unref(pb);
    gdk_pixmap_unref(pm);
    gdk_bitmap_unref(bm);
    g_free(pixfile);
  }
}

void load_resume_dialog(queue_item_t * q)
{
  GladeXML *xml;
  struct stat sbuf;
  char *filename, *p;

  g_return_if_fail(resumeDialog.window == NULL);

  xml = glade_xml_new(deadftp.gladefile, "resumeDialog");
  if (!xml) {
    g_error(_("Couldn't load the interface!"));
    return;
  }

  resumeDialog.window = glade_xml_get_widget(xml, "resumeDialog");
  STORE_WIDGET(resumeDialog, filename);

  STORE_WIDGET(resumeDialog, oldpixmap);
  STORE_WIDGET(resumeDialog, oldsize);
  STORE_WIDGET(resumeDialog, olddate);

  STORE_WIDGET(resumeDialog, newpixmap);
  STORE_WIDGET(resumeDialog, newsize);
  STORE_WIDGET(resumeDialog, newdate);

  STORE_WIDGET(resumeDialog, timeout);
  STORE_WIDGET(resumeDialog, timeout_time);

  STORE_WIDGET(resumeDialog, overwrite);
  STORE_WIDGET(resumeDialog, resume);
  STORE_WIDGET(resumeDialog, resume_all);

  glade_xml_signal_autoconnect(xml);

  gtk_object_unref(GTK_OBJECT(xml));

  memset((char *) &sbuf, 0, sizeof(sbuf));
  filename = g_strdup_printf("%s/%s", q->local_dir, q->local_file);
  stat(filename, &sbuf);
  g_free(filename);
  /* If the queue item is being sent then the filename to ask about is 
   * remote_file otherwise it's local_file */
  if (QUEUE_GET_FLAG(q, QUEUE_FLAG_SEND)) {
    gtk_label_set_text(GTK_LABEL(resumeDialog.filename), q->remote_file);

    set_file_pixmap(GTK_PIXMAP(resumeDialog.oldpixmap), q->remote_file);
    set_file_pixmap(GTK_PIXMAP(resumeDialog.newpixmap), q->local_file);

    if (q->size)
      p = g_strdup_printf(_("%ld bytes"), q->size);
    else
      p = g_strdup(_("Unknown"));
    gtk_label_set_text(GTK_LABEL(resumeDialog.oldsize), p);
    g_free(p);

    p = g_strdup_printf(_("%ld bytes"), sbuf.st_size);
    gtk_label_set(GTK_LABEL(resumeDialog.newsize), p);
    g_free(p);

    /* TODO: do dates... */
  } else {
    gtk_label_set_text(GTK_LABEL(resumeDialog.filename), q->local_file);

    set_file_pixmap(GTK_PIXMAP(resumeDialog.newpixmap), q->remote_file);
    set_file_pixmap(GTK_PIXMAP(resumeDialog.oldpixmap), q->local_file);

    if (q->size)
      p = g_strdup_printf(_("%ld bytes"), q->size);
    else
      p = g_strdup(_("Unknown"));
    gtk_label_set_text(GTK_LABEL(resumeDialog.newsize), p);
    g_free(p);

    p = g_strdup_printf(_("%ld bytes"), sbuf.st_size);
    gtk_label_set(GTK_LABEL(resumeDialog.oldsize), p);
    g_free(p);

    /* TODO: dates */
  }
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(resumeDialog.timeout),
			       deadftp_prefs.resume_timeout);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(resumeDialog.timeout_time),
			    deadftp_prefs.resume_timeout_time);

  gtk_window_set_default(GTK_WINDOW(resumeDialog.window), resumeDialog.resume);
  if (!IS_SET(FTP_FLAG_REST_SUPPORT)) {
    gtk_widget_set_sensitive(resumeDialog.resume, FALSE);
    gtk_widget_set_sensitive(resumeDialog.resume_all, FALSE);
    gtk_window_set_default(GTK_WINDOW(resumeDialog.window), resumeDialog.overwrite);
  }

  if (deadftp_prefs.resume_timeout)
    resumeDialog.timeout_tag = gtk_timeout_add(1000, queue_resume_timeout, NULL);

  gtk_widget_show(resumeDialog.window);
}

gint queue_resume_timeout(gpointer data)
{
  gint time;

  if (!resumeDialog.window)
    return FALSE;

  time =
    gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(resumeDialog.timeout_time));

  time -= 1;
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(resumeDialog.timeout_time), time);
  if (time <= 0) {
    gtk_window_activate_default(GTK_WINDOW(resumeDialog.window));
    return FALSE;
  }
  return TRUE;
}

void on_overwrite_clicked(GtkButton * button, gpointer data)
{
  QUEUE_CURRENT(FTP.queue)->resume = 0;
  QUEUE_CLEAR_FLAG(QUEUE_CURRENT(FTP.queue), QUEUE_FLAG_RESUME);

  ftp_start_data_connection(FTP_STATE_TRANS_PASV);

  resume_dialog_destroy();
}

void on_overwrite_all_clicked(GtkButton * button, gpointer data)
{
  FTP.queue.resume_action = RESUME_OVERWRITE;

  on_overwrite_clicked(button, data);
}

void on_resume_clicked(GtkButton * button, gpointer data)
{
  /* reume in the current queue item was set in ftp_check_resume */
  QUEUE_SET_FLAG(QUEUE_CURRENT(FTP.queue), QUEUE_FLAG_RESUME);

  ftp_start_data_connection(FTP_STATE_TRANS_PASV);

  resume_dialog_destroy();
}

void on_resume_all_clicked(GtkButton * button, gpointer data)
{
  FTP.queue.resume_action = RESUME_RESUME;

  on_resume_clicked(button, data);
}

void on_skip_clicked(GtkButton * button, gpointer data)
{
  QUEUE_SET_STATUS(QUEUE_CURRENT(FTP.queue), QUEUE_STATUS_PAUSE);
  queue_update_item_status(QUEUE_CURRENT(FTP.queue));

  queue_next(FALSE);

  resume_dialog_destroy();
}

void on_skip_all_clicked(GtkButton * button, gpointer data)
{
  FTP.queue.resume_action = RESUME_SKIP;

  on_skip_clicked(button, data);
}

static void resume_dialog_rename_cb(gchar * string, gpointer data)
{
  queue_item_t *q = QUEUE_CURRENT(FTP.queue);
  if (!string) {
    gtk_window_set_modal(GTK_WINDOW(resumeDialog.window), TRUE);
    return;
  }

  if (QUEUE_GET_FLAG(q, QUEUE_FLAG_SEND)) {
    strncpy(q->remote_file, string, SMALL_BUFFER);

    if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_REMOTE_FILE] != -1) {
      gtk_clist_set_text(GTK_CLIST(deadftp.queue), g_list_index(ftp_queue, q),
			 deadftp_prefs.queue_layout[QUEUE_LAYOUT_REMOTE_FILE],
			 q->remote_file);
    } else if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_REMOTE_PATH] != -1) {
      char *path = g_strdup_printf("%s/%s", q->remote_dir, q->remote_file);
      gtk_clist_set_text(GTK_CLIST(deadftp.queue), g_list_index(ftp_queue, q),
			 deadftp_prefs.queue_layout[QUEUE_LAYOUT_REMOTE_PATH], path);
      g_free(path);
    }
  } else {
    strncpy(QUEUE_CURRENT(FTP.queue)->local_file, string, SMALL_BUFFER);

    if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_LOCAL_FILE] != -1) {
      gtk_clist_set_text(GTK_CLIST(deadftp.queue), g_list_index(ftp_queue, q),
			 deadftp_prefs.queue_layout[QUEUE_LAYOUT_LOCAL_FILE],
			 q->local_file);
    } else if (deadftp_prefs.queue_layout[QUEUE_LAYOUT_LOCAL_PATH] != -1) {
      char *path = g_strdup_printf("%s/%s", q->local_dir, q->local_file);
      gtk_clist_set_text(GTK_CLIST(deadftp.queue), g_list_index(ftp_queue, q),
			 deadftp_prefs.queue_layout[QUEUE_LAYOUT_LOCAL_PATH], path);
      g_free(path);
    }
  }


  /* TODO: maybe call check_resume on the new filename ? */

  ftp_start_data_connection(FTP_STATE_TRANS_PASV);

  resume_dialog_destroy();
}

void on_rename_clicked(GtkButton * button, gpointer data)
{
  GtkWidget *dialog;

  /* Problem: The resume dialog should be modal but if it is modal then this dialog
   * doesn't get any events.  So, turn off modal on the resume dialog while this dialog
   * is open.  Is there a better way to do this ??? */
  gtk_window_set_modal(GTK_WINDOW(resumeDialog.window), FALSE);
  dialog =
    gnome_request_dialog(FALSE, _("Enter new destination filename:"), NULL, 256,
			 resume_dialog_rename_cb, NULL, NULL);
  gtk_widget_show(dialog);
}

void on_cancel_clicked(GtkButton * button, gpointer data)
{
  resume_dialog_destroy();
}

gboolean on_resumeDialog_destroy(GtkWidget * widget, GdkEvent * event,
				 gpointer user_data)
{
  resumeDialog.window = NULL;

  return FALSE;
}

void resume_dialog_destroy()
{
  gtk_timeout_remove(resumeDialog.timeout_tag);

  gtk_widget_destroy(resumeDialog.window);
}
